import json
import os
import uuid
from collections import Counter
from datetime import datetime
from filelock import FileLock
from toffee.funcov import CovGroup, get_func_full_name

from .utils import convert_line_coverage, get_toffee_custom_key_value


LOCK_FILE_FUNC_COV = "/tmp/toffee_func_cov.lock"
LOCK_FILE_LINE_COV = "/tmp/toffee_line_cov.lock"

def get_default_report_name():
    current_time = datetime.now().strftime("%Y%m%d%H%M%S")
    return f"report-{current_time}/report-{current_time}.html"


def get_template_dir():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")


__output_report_dir__ = None


def set_output_report(report_path):
    report_dir = os.path.dirname(report_path)
    if not os.path.exists(report_dir):
        os.makedirs(report_dir)

    global __output_report_dir__
    __output_report_dir__ = report_dir


def __update_line_coverage__(line_coverage_list=[], line_grate=99):
    if line_coverage_list is None:
        return None
    if len(line_coverage_list) == 0:
        return None
    (hint, all), ignore = convert_line_coverage(
        line_coverage_list, os.path.join(__output_report_dir__, "line_dat")
    )
    assert os.path.exists(
        os.path.join(__output_report_dir__, "line_dat/index.html")
    ), "Failed to convert line coverage"

    return {
        "hints": hint,
        "total": all,
        "grate": __report_info__.get("line_grate", line_grate),
        "ignore": ignore
    }


def __update_func_coverage__(__func_coverage__):
    if __func_coverage__ is None:
        return None
    coverage = {}
    group_num_hints = 0
    group_num_total = 0
    point_num_hints = 0
    point_num_total = 0
    bin_num_hints = 0
    bin_num_total = 0
    has_once = False

    """
        Data Stucture Hierachy:
            Dict(CovGroup): list[dict]  |- Points:   list[dict] |- bins:     list[dict]  |-   hints:  int
                                                                |                        |-   name:   str
                                                                |- hinted:   boolean
                                                                |- name:     str
                                                                |- functions:list[str]
                                        |- name:     str
                                        |- hinted:   boolean
                                        |- bin_num_total:  int
                                        |- bin_num_hints:  int
                                        |- point_num_total:  int
                                        |- point_num_hints:  int
                                        |- has_once:  boolean
    """

    def parse_group(g):
        nonlocal point_num_hints, point_num_total, bin_num_hints, bin_num_total, has_once
        data = json.loads(g)
        if data["has_once"]:
            has_once = True
        return data

    def merge_dicts(dict1, dict2):
        """Recursively merge two dictionaries."""
        result = dict1.copy()  # Start with keys and values of dict1
        for key, value in dict2.items():
            if key.startswith("__"):
                continue
            if key in result:
                if isinstance(result[key], dict) and isinstance(value, dict):
                    result[key] = merge_dicts(result[key], value)
                elif isinstance(result[key], list) and isinstance(value, list):
                    if key == "points" or key == "bins":
                        if key == "bins" and result.get("dynamic_bin", False) == False:
                            assert Counter([x["name"] for x in value]) == Counter(
                                x["name"] for x in result[key]
                            ), f"bins in points {dict1['name']} should be same, merge function coverage: {dict1['name']} failed (when add_watch_point, use dynamic_bin=True to ignore this error)"
                        old_keys = {a["name"]: i for i, a in enumerate(result[key])}
                        for data in value:
                            if data["name"] not in old_keys:
                                result[key].append(data)
                            else:
                                result[key][old_keys[data["name"]]] = merge_dicts(
                                    result[key][old_keys[data["name"]]], data
                                )
                    else: # Normal list merge, use unique items
                        result[key] = list(set(result[key] + value))
                elif isinstance(result[key], bool) and isinstance(value, bool):
                    if key == "has_once":
                        result[key] = result[key] and value
                    else:
                        result[key] = result[key] or value
                elif isinstance(result[key], int) and isinstance(value, int):
                    result[key] += value
                elif result[key] == value:
                    continue  # Same value, do nothing
                else:
                    raise ValueError(
                        f"Conflict at key '{key}': {result[key]} vs {value}"
                    )
            else:
                result[key] = value
        return result

    def merge_dicts_list(dicts: list):
        """Merge a list of dictionaries."""
        result = {}
        for d in dicts:
            if d["name"] in result:
                result[d["name"]] = merge_dicts(result[d["name"]], d)
            else:
                result[d["name"]] = d
        return [v for _, v in result.items()]

    coverage["groups"] = merge_dicts_list([parse_group(g) for g in __func_coverage__])

    # Recalculate the groups hinted situation
    for group in coverage["groups"]:
        group_num_total += 1

        group["hinted"] = True

        # Point Hinted
        group["point_num_total"] = len(group["points"])
        group["point_num_hints"] = 0
        group["point_functions"] = []
        group["bin_num_total"] = 0
        group["bin_num_hints"] = 0

        for point in group["points"]:
            # Bins Hinted
            group["bin_num_total"] += len(point["bins"])
            tmp_bin_hints = 0
            for bin in point["bins"]:
                group["bin_num_hints"] += int(bin["hints"] > 0)
                tmp_bin_hints += int(bin["hints"] > 0)
            point["hinted"] = tmp_bin_hints == len(point["bins"])

            group["point_num_hints"] += int(point["hinted"])
            group["hinted"] &= point["hinted"]
            group["point_functions"].extend([item for funcs in point["functions"].items() for item in funcs[1]])
        group["point_functions"] = list(set(group["point_functions"]))

        # Group Hinted
        if group["hinted"]:
            group_num_hints += 1

        point_num_hints += group["point_num_hints"]
        point_num_total += group["point_num_total"]
        bin_num_hints += group["bin_num_hints"]
        bin_num_total += group["bin_num_total"]

    coverage["group_num_total"] = group_num_total
    coverage["group_num_hints"] = group_num_hints
    coverage["point_num_total"] = point_num_total
    coverage["point_num_hints"] = point_num_hints
    coverage["bin_num_total"] = bin_num_total
    coverage["bin_num_hints"] = bin_num_hints
    coverage["grate"] = 100
    coverage["has_once"] = has_once
    return coverage


__report_info__ = {
    "user": None,
    "title": None,
    "meta": {},
}


def process_context(context, config):
    def set_ctx(key, value):
        if value is None:
            return
        context[key] = value

    for k in ["Plugins", "Packages"]:
        context["metadata"].pop(k, None)

    global_report_info = get_toffee_custom_key_value().get("toffee_report_information", {})
    set_ctx("user", global_report_info.get("user", {"name":"Tofee", "email":"-"}))
    set_ctx("title", global_report_info.get("title", "Toffee Test Report"))
    context["metadata"].update(global_report_info.get("meta", {}))

    set_ctx("user", __report_info__["user"])
    set_ctx("title", __report_info__["title"])
    context["metadata"].update(__report_info__["meta"])

    coverage_func_list = []
    coverage_func_keys = []
    coverage_line_list = []
    coverage_line_keys = []

    for t in context["tests"]:
        for p in t["phases"]:
            if hasattr(p["report"], "__coverage_group__"):
                for fc_data in p["report"].__coverage_group__:
                    key = "%s-%s" % (fc_data["hash"], fc_data["id"])
                    if key in coverage_func_keys:
                        continue
                    coverage_func_keys.append(key)
                    coverage_func_list.append(fc_data["data"])
            if hasattr(p["report"], "__line_coverage__"):
                lc_data = p["report"].__line_coverage__
                key = "%s-%s" % (lc_data["hash"], lc_data["id"])
                if key in coverage_line_keys:
                    continue
                coverage_line_keys.append(key)
                coverage_line_list.append(lc_data)
    # search data in session
    if hasattr(context["session"], "__coverage_group__"):
        for fc_data in context["session"].__coverage_group__:
            key = "%s-%s" % (fc_data["hash"], fc_data["id"])
            if key in coverage_func_keys:
                continue
            coverage_func_keys.append(key)
            coverage_func_list.append(fc_data["data"])
    if hasattr(context["session"], "__line_coverage__"):
        for lc_data in context["session"].__line_coverage__:
            key = "%s-%s" % (lc_data["hash"], lc_data["id"])
            if key in coverage_line_keys:
                continue
            coverage_line_keys.append(key)
            coverage_line_list.append(lc_data)
    context["coverages"] = {
        "line": __update_line_coverage__(coverage_line_list, global_report_info.get("line_grate", 99)),
        "functional": __update_func_coverage__(coverage_func_list),
    }
    test_abstract_info = {
        get_func_full_name(t["item"].function):t["status"]["word"] for t in context["tests"]
    }
    context["test_abstract_info"] = test_abstract_info
    if config.option.toffee_report_dump_json:
        def default(o):
            try:
                return str(o)
            except:
                return None
        with open(os.path.join(__output_report_dir__, "toffee_report.json"), "w") as f:
            json.dump(context, f, indent=4, default=default)


def set_func_coverage(request, g):
    if not isinstance(g, list):
        g = [g]
    for i in g:
        assert isinstance(
            i, CovGroup
        ), "g should be an instance of CovGroup or list of CovGroup"
    request.node.__coverage_group__ = [str(x) for x in g]
    if request.scope == 'module':
        with FileLock(LOCK_FILE_FUNC_COV):
            session = request.session
            if not hasattr(session, "__coverage_group__"):
                session.__coverage_group__ = []
            session.__coverage_group__.extend([{
                "hash": "%s" % hash(g),
                "id": "H%s-P%s" % (uuid.getnode(), os.getpid()),
                "data": g,
            } for g in request.node.__coverage_group__])


def set_line_coverage(request, datfile, ignore=[]):
    assert isinstance(datfile, str), "datfile should be a string"
    request.node.__line_coverage__ = json.dumps({"datfile":datfile, "ignore": ignore})
    if request.scope == 'module':
        with FileLock(LOCK_FILE_LINE_COV):
            session = request.session
            if not hasattr(session, "__line_coverage__"):
                session.__line_coverage__ = []
            session.__line_coverage__.append({
                "hash": "%s" % hash(datfile),
                "id": "H%s-P%s" % (uuid.getnode(), os.getpid()),
                "data": datfile,
                "ignore": ignore,
            })


def process_func_coverage(item, call, report):
    if call.when != "teardown":
        return
    if hasattr(item, "__coverage_group__"):
        groups = []
        for g in item.__coverage_group__:
            assert isinstance(
                g, str
            ), "item.__coverage_group__ should be an instance of CovGroup"
            groups.append(
                {
                    "hash": "%s" % hash(g),
                    "id": "H%s-P%s" % (uuid.getnode(), os.getpid()),
                    "data": g,
                }
            )
        report.__coverage_group__ = groups
    if hasattr(item, "__line_coverage__"):
        assert isinstance(
            item.__line_coverage__, str
        ), "item.__line_coverage__ should be a string"
        line_data = json.loads(item.__line_coverage__)
        datfile = line_data["datfile"]
        ignore = line_data["ignore"]
        report.__line_coverage__ = {
            "hash": "%s" % hash(datfile),
            "id": "H%s-P%s" % (uuid.getnode(), os.getpid()),
            "data": datfile,
            "ignore": ignore,
        }
    return report


def set_user_info(name, email):
    global __report_info__
    __report_info__["user"] = {"name": name, "email": email}


def set_title_info(title):
    global __report_info__
    __report_info__["title"] = title


def set_meta_info(key, value, is_del=False):
    global __report_info__
    if is_del:
        del __report_info__["meta"][key]
    else:
        __report_info__["meta"][key] = value


def set_line_good_rate(rate):
    global __report_info__
    __report_info__["line_grate"] = rate



import asyncio
import functools

import pytest
import pytest_asyncio
import toffee

fixture = pytest_asyncio.fixture


async def cancel_all_tasks():
    tasks = {t for t in asyncio.all_tasks() if t is not asyncio.current_task()}
    for task in tasks:
        task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)


def testcase(func):
    func.is_toffee_testcase = True

    @functools.wraps(func)
    @pytest.mark.asyncio
    async def wrapper(*args, **kwargs):
        ret = await toffee.asynchronous.main_coro(func(*args, **kwargs))
        await cancel_all_tasks()
        return ret

    return wrapper




import sys
import re
import fnmatch
from .utils import get_toffee_custom_key_value


def toffee_tags_process(item):
    '''toffee_tags(tag: Optional[list, str],         # eg: ["tag1", "tag2"], "tag1"
                   version: Optional[list,str] = [], # eg: ["v1", "v2"], "v1+", "v1-", "v1<v2"
                   skip=None                         # skip(tag, version, item): (skip, reason)
                   )
        example:
        @pytest.mark.toffee_tags(["tag1", "tag2"], ["kmh-tag-number1", "kmh-tag-number2"])
        def test_case1(...):
            pass
    '''
    import pytest
    marker = item.get_closest_marker("toffee_tags")
    kwargs = {
        "item": item,
    }
    need_args = ["tag", "version", "skip"]
    if marker:
        assert len(marker.args) < len(need_args), "Too many args, only need 3 (tag_list, version_list, skip_call_back)"
        for i, arg in enumerate(marker.args):
            kwargs[need_args[i]] = arg
        for key, value in marker.kwargs.items():
            assert key in need_args, f"Unknown key {key}"
            assert key not in kwargs, f"Duplicate args {key}"
            kwargs[key] = value        
    else:
        for arg in need_args:
            kwargs[arg] = getattr(item.module, "toffee_tags_default_%s"%arg, None)
    skip, reason = skip_process_test_tag_version(**kwargs)
    if skip:
        pytest.skip(reason)
    skip, reason = skip_process_test_cases(item.name, item.module.__name__)
    if skip:
        pytest.skip(reason)


def grep_last_number(s: str):
    m = list(re.finditer(r'(\d+(\.\d+)?)(?!.*\d)', s))
    if m:
        lm = m[-1]
        return float(lm.group(1)), s[:lm.start()]
    return None, ""


def match_version(version, version_list):
    if not version:
        return True
    version = version.strip()
    if not version_list:
        return True
    if isinstance(version_list, list):
        if len(version_list) == 0:
            return True
        return version in version_list
    assert isinstance(version_list, str), "version_list must be list or str"
    version_list = version_list.strip()
    if version in version_list:
        return True
    if "*" in version_list or "?" in version_list:
        return fnmatch.fnmatch(version, version_list)
    version_range = [-sys.maxsize, sys.maxsize]
    value, prefix = grep_last_number(version)
    # check prefix
    if "<" in version_list:
        a, b = version_list.split("<")
        if not a.strip().startswith(prefix):
            return False
        if not b.strip().startswith(prefix):
            return False
    elif not version_list.startswith(prefix):
        return False
    try:
        if "<" in version_list:
            vlist = version_list.split("<")
            assert len(vlist) == 2, "Invalid version range"
            version_range[0], _ = grep_last_number(vlist[0].strip())
            version_range[1], _ = grep_last_number(vlist[1].strip())
        else:
            target_value, _ = grep_last_number(version_list)
            if version_list.endswith("+"):
                version_range[0] = target_value
            elif version_list.endswith("-"):
                version_range[1] = target_value
            else:
                return value == target_value
    except Exception as e:
        assert False, f"Invalid version format '{version_list}', error: {e}, can not find right version number"
    return version_range[0] <= float(value) <= version_range[1]


def match_tags(source_tags, target_tags):
    tmp_normal_tags = []
    tmp_wildcard_tags = []
    if not target_tags:
        return False
    if not source_tags:
        return False
    for t in target_tags:
        t = t.strip()
        if "*" in t or "?" in t:
            tmp_wildcard_tags.append(t)
        else:
            tmp_normal_tags.append(t)
    for t in source_tags:
        t = t.strip()
        if t in tmp_normal_tags:
            return t
        for wt in tmp_wildcard_tags:
            if fnmatch.fnmatch(t, wt):
                return wt
    return False


def skip_process_test_tag_version(tag=[], version=[], skip=None, item=None):
    if isinstance(tag, str):
        tag = [tag]
    if callable(skip):
        assert item, "Case item must be provided, please dont call this function directly, use @pytest.mark.toffee_tags"
        return skip(tag, version, item)
    current_version = get_toffee_custom_key_value().get("toffee_tags_current_version", None)
    skip_tags = get_toffee_custom_key_value().get("toffee_tags_skip_tags", [])
    run_tags =  get_toffee_custom_key_value().get("toffee_tags_run_tags", [])
    if not match_version(current_version, version):
        return True, f"In Skiped version, '{current_version}' not match: '{version}'"
    tag = match_tags(tag, skip_tags)
    if tag:
        return True, f"In Skiped tags: '{tag}'"
    tag = match_tags(tag, run_tags)
    if not tag and len(run_tags) > 0:
        return True, f"No matched tags"
    return False, ""

def skip_process_test_cases(name, module):
    skip_cases = get_toffee_custom_key_value().get("toffee_tags_skip_cases", [])
    run_cases = get_toffee_custom_key_value().get("toffee_tags_run_cases", [])
    c = match_tags([name, module, "%s.%s"%(module, name)], skip_cases)
    if c:
        return True, f"In Skiped cases: '{c}'"
    c = match_tags([name, module, "%s.%s"%(module, name)], run_cases)
    if not c and len(run_cases) > 0:
        return True, f"No matched cases"
    return False, ""



import os
from zlib import crc32

from .reporter import set_func_coverage
from .reporter import set_line_coverage


class ToffeeRequest:
    def __init__(self, request):
        self.dut = None
        self.args = None
        self.request = request
        self.cov_groups = []

        self.waveform_filename = None
        self.coverage_filename = None

    def __add_cov_sample(self, cov_groups):
        """
        Add the coverage sample to the DUT.
        """

        assert self.dut is not None, "The DUT has not been set."
        assert self.cov_groups is not None, "The coverage group has not been set."

        if not isinstance(cov_groups, list):
            cov_groups = [cov_groups]

        def sample_helper(cov_point):
            return lambda _: cov_point.sample()

        for g in cov_groups:
            self.dut.xclock.StepRis(sample_helper(g))

    def __need_report(self):
        """
        Whether to generate the report
        """

        return self.request.config.getoption("--toffee-report")

    def create_dut(
        self,
        dut_cls,
        clock_name=None,
        waveform_filename=None,
        coverage_filename=None,
        dut_extra_args=(),
        dut_extra_kwargs={},
    ):
        """
        Create the DUT.

        Args:
            dut_cls: The DUT class.
            clock_name: The clock pin name.
            waveform_filename: The waveform filename. If not set, it will be auto-generated.
            coverage_filename: The coverage filename. If not set, it will be auto-generated.
            dut_extra_args: Extra positional arguments for the DUT. e.g., ("arg1", "args2")
            dut_extra_kwargs: Extra keyword arguments for the DUT. e.g., {"arg1": 1, "arg2": 2}
        Returns:
            The DUT instance.
        """

        final_kwargs = dict(dut_extra_kwargs)

        if self.__need_report():
            report_dir = os.path.dirname(self.request.config.option.report[0])
            request_name = self.request.node.name
            path_bytes = str(self.request.path).encode("utf-8")
            path_hash = format(crc32(path_bytes), 'x')
            export_filename = "_".join((dut_cls.__name__, request_name, path_hash))

            self.waveform_filename = (
                f"{report_dir}/{export_filename}.fst"
            )
            self.coverage_filename = (
                f"{report_dir}/{export_filename}.dat"
            )

            if waveform_filename is not None:
                self.waveform_filename = waveform_filename
            if coverage_filename is not None:
                self.coverage_filename = coverage_filename

            final_kwargs["waveform_filename"] = self.waveform_filename
            final_kwargs["coverage_filename"] = self.coverage_filename

            self.dut = dut_cls(*dut_extra_args, **final_kwargs)

            if self.cov_groups is not None:
                self.__add_cov_sample(self.cov_groups)
        else:
            if waveform_filename is not None:
                final_kwargs["waveform_filename"] = waveform_filename
            if coverage_filename is not None:
                final_kwargs["coverage_filename"] = coverage_filename

            self.dut = dut_cls(*dut_extra_args, **final_kwargs)

        if clock_name:
            self.dut.InitClock(clock_name)

        return self.dut

    def add_cov_groups(self, cov_groups, periodic_sample=True):
        """
        Add the coverage groups to the list.

        Args:
            cov_groups: The coverage groups to be added.
            periodic_sample: Whether to sample the coverage periodically.
        """

        if not isinstance(cov_groups, list):
            cov_groups = [cov_groups]
        self.cov_groups.extend(cov_groups)

        if self.dut is not None and periodic_sample:
            self.__add_cov_sample(cov_groups)

    def finish(self, request):
        """
        Finish the request.
        """

        if self.dut is not None:
            self.dut.Finish()

            if self.__need_report():
                set_func_coverage(request, self.cov_groups)
                set_line_coverage(request, self.coverage_filename)

        for g in self.cov_groups:
            g.clear()

        self.cov_groups.clear()


PreRequest = ToffeeRequest



import inspect
import os

import pytest
import toffee
from toffee import run

from .markers import toffee_tags_process
from .reporter import get_default_report_name
from .reporter import get_template_dir
from .reporter import process_context
from .reporter import process_func_coverage
from .reporter import set_output_report
from .utils import base64_decode
from .utils import get_toffee_custom_key_value
from .utils import set_toffee_custom_key_value


"""
toffee plugin
"""


@pytest.hookimpl(trylast=True, optionalhook=True)
def pytest_reporter_context(context, config):
    process_context(context, config)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_call(item):
    call = yield
    if call.excinfo is not None:
        eclass, evalue, _ = call.excinfo
        ignore_exceptions = get_toffee_custom_key_value().get(
            "toffee_ignore_exceptions", []
        )
        if eclass.__name__ in ignore_exceptions:
            call.force_exception(
                pytest.skip.Exception(
                    "Skiped exception: '%s(%s)'" % (eclass.__name__, evalue)
                )
            )


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()
    return process_func_coverage(item, call, report)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    toffee_tags_process(item)


def pytest_addoption(parser):
    group = parser.getgroup("reporter")
    group.addoption(
        "--toffee-report",
        action="store_true",
        default=False,
        help="Generate the report.",
    )

    group.addoption(
        "--report-name", action="store", default=None, help="The name of the report."
    )

    group.addoption(
        "--report-dir", action="store", default=None, help="The dir of the report."
    )

    group.addoption(
        "--report-dump-json",
        action="store_true",
        default=False,
        help="Dump json report.",
    )

    group.addoption(
        "--custom-key-value",
        action="store",
        default=None,
        help="Custom key value pair. dict wtih base64 encoded",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "mlvp_async: mark test to run with toffee's event loop"
    )
    config.addinivalue_line("markers", toffee_tags_process.__doc__)
    if config.getoption("--toffee-report"):
        config.option.template = ["html/toffee.html"]
        config.option.template_dir = [get_template_dir()]

        report_name = config.getoption("--report-name")
        if report_name is None:
            report_name = get_default_report_name()

        report_dir = config.getoption("--report-dir")
        if report_dir is None:
            report_dir = "reports"
        report_name = os.path.join(report_dir, report_name)

        config.option.report = [report_name]
        set_output_report(report_name)
    if config.getoption("--report-dump-json"):
        config.option.toffee_report_dump_json = True
    else:
        config.option.toffee_report_dump_json = False
    # Custom key value pair
    ckv = config.getoption("--custom-key-value")
    if ckv:
        set_toffee_custom_key_value(base64_decode(ckv))

    if "asyncio_default_fixture_loop_scope" not in config._inicache:
        config._inicache["asyncio_default_fixture_loop_scope"] = "function"


"""
toffee async test
"""


@pytest.hookimpl(tryfirst=True)
def pytest_pyfunc_call(pyfuncitem):
    if "mlvp_async" in pyfuncitem.keywords:
        toffee.warning(
            "test marked with mlvp_async will be deprecated in the future, please use \
                        @toffee_test.case instead"
        )

        func = pyfuncitem.obj
        assert inspect.iscoroutinefunction(
            func
        ), "test marked with mlvp_async must be a coroutine function"

        signature = inspect.signature(func)
        filtered_funcargs = {
            k: v for k, v in pyfuncitem.funcargs.items() if k in signature.parameters
        }

        run(func(**filtered_funcargs))

        return True

    return None


from .request import ToffeeRequest


@pytest.fixture()
def toffee_request(request):
    request_info = ToffeeRequest(request)

    yield request_info

    request_info.finish(request)


mlvp_pre_request = toffee_request



import os
import re
import subprocess
import copy
import json
import base64
import itertools
import tempfile
from typing import List, Iterable


def exe_cmd(cmd):
    if isinstance(cmd, list):
        cmd = " ".join(cmd)

    result = subprocess.run(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    success = result.returncode == 0

    stdout = result.stdout.decode("utf-8")
    stderr = result.stderr.decode("utf-8")
    return success, stdout, stderr


def parse_lines(text: str):
    pattern = r"lines\.+: \d+\.\d+% \((\d+) of (\d+) lines\)\n"
    match = re.search(pattern, text)
    if match:
        return tuple(map(int, match.groups()))
    return -1, -1


def convert_line_coverage(line_coverage_list, output_dir):
    assert isinstance(line_coverage_list, list), "Invalid line coverage list"
    coverage_info_list = []
    final_ignore_info = []
    for ldata in line_coverage_list:
        fdat = ldata["data"]
        fign = ldata["ignore"]
        if isinstance(fign, str):
            fign = [fign]
        assert isinstance(fign, list), f"Invalid data file: '{fign}'"
        assert os.path.exists(fdat), f"Invalid data file: '{fdat}'"
        su, so, se = exe_cmd(["verilator_coverage  -write-info", fdat + ".info", fdat])
        assert su, f"Failed to convert line coverage({fdat}): {se}"
        ignore_text = []
        # find all .ignore files
        ignore_file_list = []
        for f in fign:
            if os.path.isdir(f):
                for root, _, files in os.walk(f):
                    for file in files:
                        if file.endswith(".ignore"):
                            ignore_file_list.append(os.path.join(root, file))
            else:
                assert os.path.exists(f), f"Not find ignore file: '{f}'"
                ignore_file_list.append(f)
        for ignore_file in ignore_file_list:
            for i, ln in enumerate(open(ignore_file).readlines()):
                ln = ln.strip()
                if ln.startswith("#"):
                    continue
                for c in ["'", "\""]:
                    assert c not in ln, f"Invalid line number({i}): '{ln}'"
                ignore_text.append(f"\'{ln}\'")
        su, so, se = exe_cmd(
            ["lcov", "--remove", fdat + ".info", " ".join(ignore_text), "--output-file", fdat + ".info"]
        )
        assert su, f"Failed to remove line with file: '{ignore_file_list}', exception: {se}"
        coverage_info_list.append(fdat + ".info")
        final_ignore_info.append([fdat, copy.deepcopy(ignore_file_list)])
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    merged_info = os.path.join(output_dir, "merged.info")
    lcov_merge_and_del(coverage_info_list, merged_info)
    su, so, se = exe_cmd(["genhtml", "--branch-coverage", merged_info, "-o", output_dir])
    assert su, f"Failed to convert line coverage: {se}"
    return parse_lines(so), final_ignore_info


def lcov_merge_and_del(list_to_merge: List[str], out_file: str, max_kbytes=10) -> None:
    def create_temp_file() -> str:
        _, path = tempfile.mkstemp(prefix="merge_", suffix=".info")
        return path

    max_bytes = max_kbytes * 1024
    cmd_list = ["lcov", "-o", out_file, "-a", out_file]
    cmd_len = len(" ".join(cmd_list)) + 1
    temp_total_size = cmd_len
    temp_files = []
    total_size = cmd_len
    raw = 0
    cooking = 0
    for cooking in range(len(list_to_merge)):
        f = list_to_merge[cooking]
        f_len = 3 + len(f)  # The length of '-a ' is 3
        if f_len + total_size > max_bytes:
            temp = create_temp_file()  # Create temp file
            temp_files.append(temp)
            temp_total_size += len(temp) + 3
            _lcov_merge(itertools.islice(list_to_merge, raw, cooking), temp)  # Merge into the temp
            total_size = cmd_len
            raw = cooking
        total_size += f_len
    # There are still unmerged files here
    temp = create_temp_file()
    temp_files.append(temp)
    _lcov_merge(itertools.islice(list_to_merge, raw, cooking + 1), temp)
    # Delete .dat.info
    for f in list_to_merge:
        os.remove(f)
    # If the number of temp files is also huge
    if temp_total_size > max_bytes:
        lcov_merge_and_del(temp_files, out_file)
    else:
        _lcov_merge(temp_files, out_file)
    # Clean the temps
    for f in temp_files:
        os.remove(f)


def _lcov_merge(list_to_merge: Iterable[str], out_file: str) -> None:
    cmd = ["lcov", "-o", out_file]
    for f in list_to_merge:
        cmd.append("-a")
        cmd.append(f)
    subprocess.run(cmd, stdout=subprocess.PIPE, shell=False, check=True)


def base64_encode(data):
    input_bytes = json.dumps(data).encode('utf-8')
    base64_bytes = base64.b64encode(input_bytes)
    base64_str = base64_bytes.decode('utf-8')
    return base64_str


def base64_decode(base64_str):
    base64_bytes = base64_str.encode('utf-8')
    input_bytes = base64.b64decode(base64_bytes)
    return json.loads(input_bytes.decode('utf-8'))


def get_toffee_custom_key_value():
    import pytest
    return getattr(pytest, "toffee_custom_key_value", {})


def set_toffee_custom_key_value(value):
    import pytest
    assert isinstance(value, dict), "Invalid custom key value"
    pytest.toffee_custom_key_value = value



try:
    from . import __version

    __version__ = __version.version

except ImportError:
    __version__ = "unknown"

from toffee_test.testcase import testcase, fixture
from toffee_test.request import ToffeeRequest, PreRequest
from toffee_test.utils import base64_decode, base64_encode

__all__ = ["testcase", "fixture", "ToffeeRequest", "PreRequest", "base64_decode", "base64_encode"]
